// TODO: なんでもかんでもos.tsに突っ込むのやめたいのでよしなに分割する

import { Component, markRaw, Ref, ref, defineAsyncComponent } from 'vue';
import { EventEmitter } from 'eventemitter3';
import insertTextAtCursor from 'insert-text-at-cursor';
import * as foundkey from 'foundkey-js';
import { getBuiltinThemes } from './scripts/theme';
import { url } from '@/config';
import MkPostFormDialog from '@/components/post-form-dialog.vue';
import MkWaitingDialog from '@/components/waiting-dialog.vue';
import { MenuItem } from '@/types/menu';
import { $i } from '@/account';

export const pendingApiRequestsCount = ref(0);

const apiClient = new foundkey.api.APIClient({
	origin: url,
});

function convertUser(user) {
	return {
		id: user.id,
		username: user.username,
		host: user.fqn.split('@')[1],
		avatarUrl: user.avatar,
		avatarBlurhash: '',
		isAdmin: user.pleroma?.is_admin || false,
		isModerator: user.pleroma?.is_moderator || false,
		isBot: user.bot,
		isCat: true, // TODO lmao
		instance: {
			name: user.akkoma.instance.name,
			softwareName: user.akkoma.instance.nodeinfo.software.name,
			softwareVersion: user.akkoma.instance.nodeinfo.software.version,
			iconUrl: user.akkoma.instance.favicon,
			faviconUrl: user.akkoma.instance.favicon,
			themeColor: 'transparent',
		},
		emojis: [], // TODO
		onlineStatus: 'unknown',
		url: user.url,
		uri: user.url,
		createdAt: user.created_at,
		updatedAt: user.last_status_at,
		lastFetchedAt: user.last_status_at,
		bannerUrl: user.header,
		bannerBlurhash: '',
		isLocked: user.locked,
		isSilenced: false,
		isSuspended: false,
		description: user.source.note,
		location: null,
		birthday: null,
		lang: null,
		fields: [], // TODO
		followersCount: user.followers_count,
		followingCount: user.following_count,
		notesCount: user.statuses_count,
		pinnedNoteIds: [],
		pinnedNotes: [],
		pinnedPageId: null,
		pinnedPage: null,
		publicReactions: false,
		ffVisibility: 'public',
		twoFactorEnabled: false,
		usePasswordLessLogin: false,
		securityKeys: false,
	};
}

function convertVisibility(visibility: string): string {
	return { public: 'public', unlisted: 'home', private: 'followers', direct: 'specified' }[visibility];
}

function convertNote(note) {
	return {
		id: note.id,
		createdAt: note.created_at,
		userId: note.account.id,
		user: convertUser(note.account),
		text: note.akkoma?.source?.content,
		cw: note.spoiler_text ? note.spoiler_text : null,
		visibility: convertVisibility(note.visibility),
		renoteCount: note.reblogs_count,
		repliesCount: note.replies_count,
		reactions: {}, // TODO,
		emojis: [], // TODO,
		fileIds: [], // TODO
		files: [], // TODO
		replyId: note.in_reply_to_id,
		renoteId: note.reblog?.id,
		mentions: note.mentions.map(mention => mention.id),
		uri: note.url,
		reply: null,
	};
}

export const api = (async (endpoint: string, args: Record<string, any> = {}, token?: string | null | undefined) => {
	pendingApiRequestsCount.value++;

	try {
		const authorizationToken = token ?? $i?.token ?? undefined;
		const authorization = authorizationToken ? `Bearer ${authorizationToken}` : undefined;

		switch (endpoint) {
			// not implemented area
			case 'get-online-users-count':
				return { count: 0 };
			case 'hashtags/list':
			case 'notes/featured':
			case 'announcements':
				return [];
			case 'charts/user/notes':
				return { total: [], inc: [], dec: [], diffs: { normal: [], "reply": [], "renote": [], "withFile": [] } };

			// implemented area
			case 'users/notes': {
				const params = new URLSearchParams({});
				if (args.limit) params.set("limit", args.limit);
				const resp = await fetch(`${url}/api/v1/accounts/${args.userId}/statuses?${params.toString()}`);
				if (resp.status !== 200) {
					throw await resp.json();
				}
				const notes = await resp.json();
				return notes.map(convertNote);
			}
			case 'users/show': {
				const acct = args.username + (args.host !== null ? `@${args.host}` : '');
				const params = new URLSearchParams({ acct });
				const resp = await fetch(`${url}/api/v1/accounts/lookup?${params.toString()}`);
				if (resp.status !== 200) {
					throw await resp.json();
				}
				const user = await resp.json();
				return convertUser(user);
			}
			case 'meta': {
				const resp = await fetch(`${url}/api/v1/instance`);
				if (resp.status !== 200) {
					throw await resp.json();
				}
				const meta = await resp.json();
				const builtinThemes = await getBuiltinThemes();
				return {
					'maintainerName': meta.email,
					'maintainerEmail': meta.email,
					'version': meta.version,
					'name': meta.title,
					'uri': meta.uri,
					'description': meta.description,
					'langs': meta.languages,
					'tosUrl': '',
					'defaultDarkTheme': builtinThemes.find(th => th.name === 'Mi Cherry Dark'),
					'defaultLightTheme': builtinThemes.find(th => th.name === 'Mi Cherry Light'),
					'disableRegistration': !meta.registrations,
					'disableLocalTimeline': false,
					'disableGlobalTimeline': false,
					'driveCapacityPerLocalUserMb': 0,
					'driveCapacityPerRemoteUserMb': 0,
					'cacheRemoteFiles': true,
					'emailRequiredForSignup': true,
					'enableHcaptcha': false,
					// "hcaptchaSiteKey": "string",
					'enableRecaptcha': false,
					// "recaptchaSiteKey": "string",
					// "swPublickey": "string",
					'bannerUrl': 'string',
					'iconUrl': meta.thumbnail,
					'maxNoteTextLength': meta.max_toot_chars || 500,
					// "emojis": [
					// 	{
					// 		"id": "string",
					// 		"aliases": [
					// 			"string"
					// 		],
					// 		"category": "string",
					// 		"host": "string",
					// 		"url": "string"
					// 	}
					// ],
					'requireSetup': false,
					'enableEmail': true,
					'translatorAvailable': false,
					// "proxyAccountName": "string",
					// "images": {
					// 	"info": "string",
					// 	"notFound": "string",
					// 	"error": "string"
					// },
					'features': {
						'registration': meta.registrations,
						'localTimeLine': true,
						'globalTimeLine': true,
						'elasticsearch': false,
						'hcaptcha': false,
						'recaptcha': false,
						'objectStorage': false,
						'serviceWorker': false,
						'miauth': false,
					},
				};
			}
			default:
				console.error('unhandled api', { endpoint, args });
		}
	} finally {
		pendingApiRequestsCount.value--;
	}
}) as typeof apiClient.request;

export const apiGet = api;

export const apiWithDialog = ((
	endpoint: string,
	data: Record<string, any> = {},
	token?: string | null | undefined,
) => {
	debugger;
	const promise = api(endpoint, data, token);
	promiseDialog(promise, null, (err) => {
		alert({
			type: 'error',
			text: (err.message + '\n' + (err.endpoint ?? '') + ' ' + (err.code ?? '')).trim(),
		});
	});

	return promise;
}) as typeof api;

export function promiseDialog<T extends Promise<any>>(
	promise: T,
	onSuccess?: ((res: any) => void) | null,
	onFailure?: ((err: Error) => void) | null,
	text?: string,
): T {
	const showing = ref(true);
	const success = ref(false);

	promise.then(res => {
		if (onSuccess) {
			showing.value = false;
			onSuccess(res);
		} else {
			success.value = true;
			window.setTimeout(() => {
				showing.value = false;
			}, 1000);
		}
	}).catch(err => {
		showing.value = false;
		if (onFailure) {
			onFailure(err);
		} else {
			alert({
				type: 'error',
				text: (err.message + '\n' + (err?.endpoint ?? '') + (err?.code ?? '')).trim(),
			});
		}
	});

	// NOTE: dynamic import results in strange behaviour (showing is not reactive)
	popup(MkWaitingDialog, {
		success,
		showing,
		text,
	}, {}, 'closed');

	return promise;
}

let popupIdCount = 0;
export const popups = ref([]) as Ref<{
	id: any;
	component: any;
	props: Record<string, any>;
	events: any[];
}[]>;

const zIndexes = {
	low: 1000000,
	middle: 2000000,
	high: 3000000,
};
export function claimZIndex(priority: 'low' | 'middle' | 'high' = 'low'): number {
	zIndexes[priority] += 100;
	return zIndexes[priority];
}

export async function popup(component: Component, props: Record<string, any>, events = {}, disposeEvent?: string) {
	markRaw(component);

	const id = ++popupIdCount;
	const dispose = () => {
		// このsetTimeoutが無いと挙動がおかしくなる(autocompleteが閉じなくなる)。Vueのバグ？
		window.setTimeout(() => {
			popups.value = popups.value.filter(popup => popup.id !== id);
		}, 0);
	};
	const state = {
		component,
		props,
		events: disposeEvent ? {
			...events,
			[disposeEvent]: dispose,
		} : events,
		id,
	};

	popups.value.push(state);

	return {
		dispose,
	};
}

export function pageWindow(path: string) {
	popup(defineAsyncComponent(() => import('@/components/page-window.vue')), {
		initialPath: path,
	}, {}, 'closed');
}

export function modalPageWindow(path: string) {
	popup(defineAsyncComponent(() => import('@/components/modal-page-window.vue')), {
		initialPath: path,
	}, {}, 'closed');
}

export function alert(props: {
	type?: 'error' | 'info' | 'success' | 'warning' | 'waiting' | 'question';
	title?: string | null;
	text?: string | null;
}): Promise<void> {
	return new Promise((resolve) => {
		popup(defineAsyncComponent(() => import('@/components/dialog.vue')), props, {
			done: resolve,
		}, 'closed');
	});
}

export function confirm(props: {
	type: 'error' | 'info' | 'success' | 'warning' | 'waiting' | 'question';
	title?: string | null;
	text?: string | null;
}): Promise<{ canceled: boolean }> {
	return new Promise((resolve) => {
		popup(defineAsyncComponent(() => import('@/components/dialog.vue')), {
			...props,
			showCancelButton: true,
		}, {
			done: result => {
				resolve(result ? result : { canceled: true });
			},
		}, 'closed');
	});
}

export function inputText(props: {
	type?: 'text' | 'email' | 'password' | 'url';
	title?: string | null;
	text?: string | null;
	placeholder?: string | null;
	default?: string | null;
}): Promise<{ canceled: true; result: undefined; } | {
	canceled: false; result: string;
}> {
	return new Promise((resolve) => {
		popup(defineAsyncComponent(() => import('@/components/dialog.vue')), {
			title: props.title,
			text: props.text,
			input: {
				type: props.type,
				placeholder: props.placeholder,
				default: props.default ?? '',
			},
		}, {
			done: result => {
				resolve(result ? result : { canceled: true });
			},
		}, 'closed');
	});
}

export function inputNumber(props: {
	title?: string | null;
	text?: string | null;
	placeholder?: string | null;
	default?: number | null;
}): Promise<{ canceled: true; result: undefined; } | {
	canceled: false; result: number;
}> {
	return new Promise((resolve) => {
		popup(defineAsyncComponent(() => import('@/components/dialog.vue')), {
			title: props.title,
			text: props.text,
			input: {
				type: 'number',
				placeholder: props.placeholder,
				default: props.default,
			},
		}, {
			done: result => {
				resolve(result ? result : { canceled: true });
			},
		}, 'closed');
	});
}

export function inputDate(props: {
	title?: string | null;
	text?: string | null;
	placeholder?: string | null;
	default?: Date | null;
}): Promise<{ canceled: true; result: undefined; } | {
	canceled: false; result: Date;
}> {
	return new Promise((resolve) => {
		popup(defineAsyncComponent(() => import('@/components/dialog.vue')), {
			title: props.title,
			text: props.text,
			input: {
				type: 'date',
				placeholder: props.placeholder,
				default: props.default,
			},
		}, {
			done: result => {
				resolve(result ? { result: new Date(result.result), canceled: false } : { canceled: true });
			},
		}, 'closed');
	});
}

export function select<C = any>(props: {
	title?: string | null;
	text?: string | null;
	default?: string | null;
} & ({
	items: {
		value: C;
		text: string;
	}[];
} | {
	groupedItems: {
		label: string;
		items: {
			value: C;
			text: string;
		}[];
	}[];
})): Promise<{ canceled: true; result: undefined; } | {
	canceled: false; result: C;
}> {
	return new Promise((resolve) => {
		popup(defineAsyncComponent(() => import('@/components/dialog.vue')), {
			title: props.title,
			text: props.text,
			select: {
				items: props.items,
				groupedItems: props.groupedItems,
				default: props.default,
			},
		}, {
			done: result => {
				resolve(result ? result : { canceled: true });
			},
		}, 'closed');
	});
}

export function success() {
	return new Promise((resolve) => {
		const showing = ref(true);
		window.setTimeout(() => {
			showing.value = false;
		}, 1000);
		popup(defineAsyncComponent(() => import('@/components/waiting-dialog.vue')), {
			success: true,
			showing,
		}, {
			done: () => resolve(),
		}, 'closed');
	});
}

export function waiting() {
	return new Promise((resolve) => {
		const showing = ref(true);
		popup(defineAsyncComponent(() => import('@/components/waiting-dialog.vue')), {
			success: false,
			showing,
		}, {
			done: () => resolve(),
		}, 'closed');
	});
}

export function form(title, form) {
	return new Promise((resolve) => {
		popup(defineAsyncComponent(() => import('@/components/form-dialog.vue')), { title, form }, {
			done: result => {
				resolve(result);
			},
		}, 'closed');
	});
}

export async function selectUser() {
	return new Promise((resolve) => {
		popup(defineAsyncComponent(() => import('@/components/user-select-dialog.vue')), {}, {
			ok: user => {
				resolve(user);
			},
		}, 'closed');
	});
}

export async function selectDriveFile(multiple: boolean) {
	return new Promise((resolve) => {
		popup(defineAsyncComponent(() => import('@/components/drive-select-dialog.vue')), {
			type: 'file',
			multiple,
		}, {
			done: files => {
				if (files) {
					resolve(multiple ? files : files[0]);
				}
			},
		}, 'closed');
	});
}

export async function selectDriveFolder(multiple: boolean) {
	return new Promise((resolve) => {
		popup(defineAsyncComponent(() => import('@/components/drive-select-dialog.vue')), {
			type: 'folder',
			multiple,
		}, {
			done: folders => {
				if (folders) {
					resolve(multiple ? folders : folders[0]);
				}
			},
		}, 'closed');
	});
}

export async function pickEmoji(src: HTMLElement | null, opts) {
	return new Promise((resolve) => {
		popup(defineAsyncComponent(() => import('@/components/emoji-picker-dialog.vue')), {
			src,
			...opts,
		}, {
			done: emoji => {
				resolve(emoji);
			},
		}, 'closed');
	});
}

export async function cropImage(image: foundkey.entities.DriveFile, options: {
	aspectRatio: number;
}): Promise<foundkey.entities.DriveFile> {
	return new Promise((resolve) => {
		popup(defineAsyncComponent(() => import('@/components/cropper-dialog.vue')), {
			file: image,
			aspectRatio: options.aspectRatio,
		}, {
			ok: x => {
				resolve(x);
			},
		}, 'closed');
	});
}

type AwaitType<T> =
	T extends Promise<infer U> ? U :
	T extends (...args: any[]) => Promise<infer V> ? V :
	T;
let openingEmojiPicker: AwaitType<ReturnType<typeof popup>> | null = null;
let activeTextarea: HTMLTextAreaElement | HTMLInputElement | null = null;
export async function openEmojiPicker(src?: HTMLElement, opts, initialTextarea: typeof activeTextarea) {
	if (openingEmojiPicker) return;

	activeTextarea = initialTextarea;

	const textareas = document.querySelectorAll('textarea, input');
	for (const textarea of Array.from(textareas)) {
		textarea.addEventListener('focus', () => {
			activeTextarea = textarea;
		});
	}

	const observer = new MutationObserver(records => {
		for (const record of records) {
			for (const node of Array.from(record.addedNodes).filter(node => node instanceof HTMLElement) as HTMLElement[]) {
				const textareas = node.querySelectorAll('textarea, input') as NodeListOf<NonNullable<typeof activeTextarea>>;
				for (const textarea of Array.from(textareas).filter(textarea => textarea.dataset.preventEmojiInsert == null)) {
					if (document.activeElement === textarea) activeTextarea = textarea;
					textarea.addEventListener('focus', () => {
						activeTextarea = textarea;
					});
				}
			}
		}
	});

	observer.observe(document.body, {
		childList: true,
		subtree: true,
		attributes: false,
		characterData: false,
	});

	openingEmojiPicker = await popup(defineAsyncComponent(() => import('@/components/emoji-picker-window.vue')), {
		src,
		...opts,
	}, {
		chosen: emoji => {
			insertTextAtCursor(activeTextarea, emoji);
		},
		closed: () => {
			openingEmojiPicker!.dispose();
			openingEmojiPicker = null;
			observer.disconnect();
		},
	});
}

export function popupMenu(items: MenuItem[] | Ref<MenuItem[]>, src?: HTMLElement, options?: {
	align?: string;
	width?: number;
	viaKeyboard?: boolean;
}) {
	return new Promise((resolve) => {
		let dispose;
		popup(defineAsyncComponent(() => import('@/components/ui/popup-menu.vue')), {
			items,
			src,
			width: options?.width,
			align: options?.align,
			viaKeyboard: options?.viaKeyboard,
		}, {
			closed: () => {
				resolve();
				dispose();
			},
		}).then(res => {
			dispose = res.dispose;
		});
	});
}

export function contextMenu(items: MenuItem[] | Ref<MenuItem[]>, ev: MouseEvent) {
	ev.preventDefault();
	return new Promise((resolve) => {
		let dispose;
		popup(defineAsyncComponent(() => import('@/components/ui/context-menu.vue')), {
			items,
			ev,
		}, {
			closed: () => {
				resolve();
				dispose();
			},
		}).then(res => {
			dispose = res.dispose;
		});
	});
}

export function post(props: Record<string, any> = {}) {
	return new Promise((resolve) => {
		// NOTE: MkPostFormDialogをdynamic importするとiOSでテキストエリアに自動フォーカスできない
		// NOTE: ただ、dynamic importしない場合、MkPostFormDialogインスタンスが使いまわされ、
		//       Vueが渡されたコンポーネントに内部的に__propsというプロパティを生やす影響で、
		//       複数のpost formを開いたときに場合によってはエラーになる
		//       もちろん複数のpost formを開けること自体Misskeyサイドのバグなのだが
		let dispose;
		popup(MkPostFormDialog, props, {
			closed: () => {
				resolve();
				dispose();
			},
		}).then(res => {
			dispose = res.dispose;
		});
	});
}

export const deckGlobalEvents = new EventEmitter();

/*
export function checkExistence(fileData: ArrayBuffer): Promise<any> {
	return new Promise((resolve) => {
		const data = new FormData();
		data.append('md5', getMD5(fileData));

		os.api('drive/files/find-by-hash', {
			md5: getMD5(fileData)
		}).then(resp => {
			resolve(resp.length > 0 ? resp[0] : null);
		});
	});
}*/
